/* TileRasterCache, a library to download tiles from several sources.
 *
 * Copyright (C) 2010 Prodevelop SL.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *  
 *   Prodevelop Integraci�n de Tecnolog�as SL
 *   Plaza don Juan de Vilarrasa , 14-5
 *   46000 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   gis@prodevelop.es
 *   http://www.prodevelop.es
 *   
 *   author: Alberto Romeu Carrasco (aromeu@prodevelop.es)
 */

package es.prodevelop.tilecache.generator;

import java.util.logging.Level;
import java.util.logging.Logger;

import es.prodevelop.gvsig.mini.common.CompatManager;
import es.prodevelop.gvsig.mini.common.IBitmap;
import es.prodevelop.gvsig.mini.common.IContext;
import es.prodevelop.gvsig.mini.common.IHandler;
import es.prodevelop.gvsig.mini.common.IRect;
import es.prodevelop.gvsig.mini.common.impl.BaseEvent;
import es.prodevelop.gvsig.mini.common.impl.Tile;
import es.prodevelop.gvsig.mini.exceptions.BaseException;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.geom.Pixel;
import es.prodevelop.gvsig.mini.geom.Point;
import es.prodevelop.gvsig.mini.map.GeoMath;
import es.prodevelop.gvsig.mini.map.GeoUtils;
import es.prodevelop.gvsig.mini.map.ViewPort;
import es.prodevelop.gvsig.mini.projection.TileConversor;
import es.prodevelop.gvsig.mini.utiles.Cancellable;
import es.prodevelop.gvsig.mini.utiles.ThreadPool;
import es.prodevelop.gvsig.mini.utiles.WorkQueue;
import es.prodevelop.tilecache.provider.TileProvider;
import es.prodevelop.tilecache.provider.filesystem.strategy.ITileFileSystemStrategy;
import es.prodevelop.tilecache.renderer.MapRenderer;

/**
 * The main class of the download process. Basically it iterates an Extent
 * obtaining all the tile numbers and executing on a Thread pool a task to
 * download/save the tiles from the server.
 * 
 * It applies several strategies and procedures to make the process more
 * flexible.
 * 
 * @author aromeu
 * 
 */
public class TileGenerator implements GeoUtils {

	private final static Logger log = Logger.getLogger("TileGenerator");
	private MapRenderer mapRenderer;
	private int mapWidth;
	private int mapHeight;
	private int fromZoomLevel;
	private int toZoomLevel;

	private int centerPixelX;
	private int centerPixelY;

	private TileProvider mTileProvider;
	private IContext context;

	private Cancellable cancellable;
	private Extent extent;
	private boolean downloadAllExtent = false;
	private IBufferStrategy bufferStrategy;
	private ITileSorter tileSorter;
	private ITileBufferIntersector tileBufferIntersector;
	private IHandler callBackHandler;

	/**
	 * The constructor
	 * 
	 * @param context
	 *            The IContext implementation
	 * @param aMapRenderer
	 *            A valid MapRenderer @see Layers
	 * @param fromZoomLevel
	 *            minimum zoom level from where download tiles
	 * @param toZoomLevel
	 *            maximum zoom level to where download tiles
	 * @param cancellable
	 *            A Cancellable instance. On each iteration the Cancellable
	 *            state is checked to see if the process has to be stopped
	 * @param extent
	 *            The extent to be downloaded
	 * @param sorter
	 *            An ITileSorter implementation. If this method is not used to
	 *            feed a map viewer, this parameter can(must) be null
	 * @param callBackHandler
	 *            An IDownloadCallBackHandler implementation. TileProvider
	 *            should notify this handler that internally will notify an
	 *            IDownloadWaiter implementation
	 * @param intersector
	 *            An ITileBufferIntersector. This parama allows the library to
	 *            download only the tiles that intersects with a single polygon.
	 *            If the library is used to feed a map viewer where the area to
	 *            download is always an Extent this paramater can be null
	 * @param mode
	 *            MODE_ONLINE = 0 - Gets tiles from file system and from the
	 *            server MODE_OFFLINE = 1 - Only gets tiles that are on file
	 *            system MODE_UPDATE = 2 - If a tile is on the file system,
	 *            deletes it and tries to download MODE_DELETE = 3 - Deletes the
	 *            tile cache MODE_CONNECTED = 4 - Don't save any tile, always
	 *            download them
	 * @param fsStrategy
	 *            The ITileFileSystemStrategy
	 * @param bufferStrategy
	 *            The IBufferStrategy
	 * @param notLocalCache if false not stores tiles in memory cache
	 * @throws BaseException
	 *             A basic check of the properties is done. If zoom levels are
	 *             invalid or the area to download is zero, then a BaseException
	 *             * is thrown
	 */
	public TileGenerator(IContext context, MapRenderer aMapRenderer,
			int fromZoomLevel, int toZoomLevel, Cancellable cancellable,
			Extent extent, ITileSorter sorter, IHandler callBackHandler,
			ITileBufferIntersector intersector, int mode,
			ITileFileSystemStrategy fsStrategy, IBufferStrategy bufferStrategy, boolean notLocalCache)
			throws BaseException {
		this.mTileProvider = new TileProvider(context, callBackHandler, 1, 1,
				aMapRenderer.getMAPTILE_SIZEPX(), mode, fsStrategy, notLocalCache);
		this.bufferStrategy = bufferStrategy;
		this.mapRenderer = aMapRenderer;
		this.fromZoomLevel = fromZoomLevel;
		this.toZoomLevel = toZoomLevel;
		this.tileBufferIntersector = intersector;
		this.tileSorter = sorter;

		this.cancellable = cancellable;
		this.extent = (Extent) extent.clone();
		downloadAllExtent = true;
		this.callBackHandler = callBackHandler;

		this.recalcWidthHeightOfAreaToDownload(fromZoomLevel);

		checkParams();

		context.sendMessage(callBackHandler, new BaseEvent(String.valueOf(this
				.calculateTotalNumTilesToDownload())),
				GeoUtils.TILEPROVIDER_TOTAL_TILES);
	}

	// TODO: Javadoc
	public TileGenerator(IContext context, MapRenderer aMapRenderer,
			ITileSorter sorter, IHandler callBackHandler,
			ITileBufferIntersector intersector, int mode,
			ITileFileSystemStrategy fsStrategy, IBufferStrategy bufferStrategy,
			int mapWidth, int mapHeight) throws BaseException {
		this.mTileProvider = new TileProvider(context, callBackHandler,
				mapWidth, mapHeight, aMapRenderer.getMAPTILE_SIZEPX(), mode,
				fsStrategy);
		this.bufferStrategy = bufferStrategy;
		this.mapRenderer = aMapRenderer;
		this.tileBufferIntersector = intersector;
		this.tileSorter = sorter;
		downloadAllExtent = true;
		this.callBackHandler = callBackHandler;
		this.mapHeight = mapHeight;
		this.mapWidth = mapWidth;
		this.centerPixelX = (int) (mapWidth / 2);
		this.centerPixelY = (int) (mapHeight / 2);
	}

	private void recalcWidthHeightOfAreaToDownload(int zoomLevel) {
		double eWidth = extent.getWidth();
		double eHeight = extent.getHeight();
		mapWidth = Math.max(
				(int) (eWidth / mapRenderer.resolutions[zoomLevel]), 1);
		mapHeight = Math.max(
				(int) (eHeight / mapRenderer.resolutions[zoomLevel]), 1);

		this.centerPixelX = (int) (mapWidth / 2);
		this.centerPixelY = (int) (mapHeight / 2);

		ViewPort.mapHeight = mapHeight;
		ViewPort.mapWidth = mapWidth;
	}

	private void checkParams() throws BaseException {
		if (fromZoomLevel > toZoomLevel)
			throw new BaseException("Error: FromZoomlevel " + fromZoomLevel
					+ " > toZoomLevel " + toZoomLevel);

		if (mapWidth == 0 || mapHeight == 0)
			throw new BaseException("Error width or height == 0: mapWidth "
					+ mapWidth + " mapHeight " + mapHeight);
	}

	private boolean checkHasBeenCancelled() {
		final boolean cancelled = TileGenerator.this.cancellable.getCanceled();
		if (cancelled) {
			TileGenerator.this.mTileProvider.clearPendingQueue();
			WorkQueue.getInstance().clearPendingTasks();
			context.sendMessage(callBackHandler, new BaseEvent(
					"Tile provider canceled"), GeoUtils.TILEPROVIDER_CANCELED);
		}
		return cancelled;
	}

	// TODO: Javadoc
	public IBitmap getMapTile(final Tile tile, final Cancellable cancellable) {
		return this.mTileProvider.getMapTileFromCache(tile, cancellable);
	}

	// TODO: Javadoc
	public void downloadExtentAsynch(final int zoomLevel, final Extent extent,
			Cancellable cancellable) {

		this.cancellable = cancellable;
		this.extent = extent;

		ThreadPool.getInstance(1).assign(new Runnable() {

			public void run() {
				try {
					// long t1 = System.currentTimeMillis();
					checkHasBeenCancelled();

					final MapRenderer renderer = mapRenderer;
					// renderer.setZoomLevel(zoomLevel);
					final int tileSizePx = renderer.getMAPTILE_SIZEPX();

					final int[] centerMapTileCoords = renderer
							.getMapTileFromCenter();

					final int additionalTilesNeededToLeftOfCenter;
					final int additionalTilesNeededToRightOfCenter;
					final int additionalTilesNeededToTopOfCenter;
					final int additionalTilesNeededToBottomOfCenter;

					final Pixel upperLeftCornerOfCenterMapTile = renderer
							.getUpperLeftCornerOfCenterMapTileInScreen(null);

					final int centerMapTileScreenLeft = upperLeftCornerOfCenterMapTile
							.getX();
					final int centerMapTileScreenTop = upperLeftCornerOfCenterMapTile
							.getY();

					final int centerMapTileScreenRight = centerMapTileScreenLeft
							+ tileSizePx;
					final int centerMapTileScreenBottom = centerMapTileScreenTop
							+ tileSizePx;

					int[] additionalTiles = bufferStrategy
							.calculateBufferAroundCenterTile(
									(int) ViewPort.mapWidth,
									(int) ViewPort.mapHeight, tileSizePx,
									centerMapTileScreenLeft,
									centerMapTileScreenRight,
									centerMapTileScreenTop,
									centerMapTileScreenBottom);

					additionalTilesNeededToLeftOfCenter = additionalTiles[0];
					additionalTilesNeededToRightOfCenter = additionalTiles[1];
					additionalTilesNeededToTopOfCenter = additionalTiles[2];
					additionalTilesNeededToBottomOfCenter = additionalTiles[3];

					// final int mapTileUpperBound = bufferStrategy
					// .calculateTileUpperBound(zoomLevel);
					final int[] mapTileCoords = new int[] {
							centerMapTileCoords[MAPTILE_LATITUDE_INDEX],
							centerMapTileCoords[MAPTILE_LONGITUDE_INDEX] };

					final int size = (additionalTilesNeededToBottomOfCenter
							+ additionalTilesNeededToTopOfCenter + 1)
							* (additionalTilesNeededToRightOfCenter
									+ additionalTilesNeededToLeftOfCenter + 1);
					Tile[] tiles = new Tile[size];
					int cont = 0;

					final Extent maxExtent = renderer.getExtent();

					final Extent viewExtent = ViewPort.calculateExtent(renderer
							.getCenter(), renderer.resolutions[zoomLevel],
							ViewPort.mapWidth, ViewPort.mapHeight);

					double[] coords = new double[] {
							maxExtent.getLefBottomCoordinate().getX(),
							maxExtent.getLefBottomCoordinate().getY() };
					final int[] leftBottom = renderer.toPixels(coords);

					coords = new double[] {
							maxExtent.getRightTopCoordinate().getX(),
							maxExtent.getRightTopCoordinate().getY() };
					final int[] rightTop = renderer.toPixels(coords);

					IRect r = TileGenerator.this.getContext().getRectangle();
					r.set(leftBottom[0], rightTop[1], rightTop[0],
							leftBottom[1]);

					boolean process = true;

					for (int y = -additionalTilesNeededToTopOfCenter; y <= additionalTilesNeededToBottomOfCenter; y++) {
						for (int x = -additionalTilesNeededToLeftOfCenter; x <= additionalTilesNeededToRightOfCenter; x++) {
							process = true;
							checkHasBeenCancelled();

							if (viewExtent.intersect(maxExtent)) {
								final int tileLeft = ViewPort.mTouchMapOffsetX
										+ centerMapTileScreenLeft
										+ (x * tileSizePx);
								final int tileTop = ViewPort.mTouchMapOffsetY
										+ centerMapTileScreenTop
										+ (y * tileSizePx);

								process &= r.intersects(tileLeft, tileTop,
										tileLeft + tileSizePx / 2, tileTop
												+ tileSizePx / 2);
							}

							if (process) {

								mapTileCoords[MAPTILE_LATITUDE_INDEX] = centerMapTileCoords[MAPTILE_LATITUDE_INDEX]
										+ renderer.isTMS() * y;

								mapTileCoords[MAPTILE_LONGITUDE_INDEX] = centerMapTileCoords[MAPTILE_LONGITUDE_INDEX]
										+ x;

								final String tileURLString = renderer
										.getTileURLString(mapTileCoords,
												zoomLevel);

								final int[] tile = new int[] {
										mapTileCoords[0], mapTileCoords[1] };

								final int tileLeft = ViewPort.mTouchMapOffsetX
										+ centerMapTileScreenLeft
										+ (x * tileSizePx);
								final int tileTop = ViewPort.mTouchMapOffsetY
										+ centerMapTileScreenTop
										+ (y * tileSizePx);

								final Tile t = new Tile(tileURLString, tile,
										new Pixel(tileLeft, tileTop),
										zoomLevel, renderer.getNAME(),
										TileGenerator.this.cancellable, null);
								if (tileSorter == null) {
									TileGenerator.this.mTileProvider
											.getMapTile(
													t.clone(),
													TileGenerator.this.cancellable);
								} else {
									if (cont < tiles.length)
										tiles[cont] = t;
								}
							} else {
								context.sendMessage(callBackHandler,
										new BaseEvent("Tile skipped"),
										GeoUtils.TILEPROVIDER_SKIPPED);
							}
							cont++;
						}
					}

					if (tileSorter != null) {
						tileSorter.sortTiles(tiles,
								TileGenerator.this.centerPixelX,
								TileGenerator.this.centerPixelY, tileSizePx);
						final int length = tiles.length;
						Tile temp;
						for (int j = 0; j < length; j++) {
							temp = tiles[j];
							if (temp != null) {
								// This should notify the callback Handlers when
								// the
								// tile is
								// cached
								TileGenerator.this.mTileProvider.getMapTile(
										temp.clone(),
										TileGenerator.this.cancellable);
							}
						}
					}
					tiles = null;
					// System.out.println("Time to request tiles: " +
					// (System.currentTimeMillis() - t1));
				} catch (Exception e) {
					log.log(Level.FINE, "onDraw", e);
				}
			}
		});
	}

	/**
	 * Iterates all the tiles from a level (given an Extent) and launches a task
	 * on a thread pool to download/save the tile
	 * 
	 * First of all calculates the number of tiles to be downlaoded. A
	 * IBufferStrategy is applied in order to allow to download more tiles than
	 * those that fit the Extent
	 * 
	 * Once the total number of tiles has been retrieved, it makes an iteration
	 * for all rows and columns (x and y values), for each iteration checks if
	 * the extent of the tile to be downloaded intersects with the extent of the
	 * view and applies the ITileBufferIntersector to see if the extent of the
	 * tile intersects with an arbitrary geometry
	 * 
	 * If the tile must be downloaded then the MapRenderer gives the URL to
	 * download and the getMapTile method of TileProvider is called.
	 * 
	 * It also checks if an ITileSorter has been set, so the tiles will be
	 * downloaded on a specific order
	 * 
	 * @param zoomLevel
	 *            The zoom level to download
	 * @param extent
	 *            The extent to download
	 * @param cancellable
	 *            A Cancellable instance to stop the process
	 */
	public void downloadLevel(final int zoomLevel, final Cancellable cancellable) {
		try {

			final MapRenderer renderer = mapRenderer;
			renderer.setZoomLevel(zoomLevel);
			final int tileSizePx = renderer.getMAPTILE_SIZEPX();

			if (this.downloadAllExtent) {
				this.recalcWidthHeightOfAreaToDownload(zoomLevel);
			}

			int[] centerMapTileCoords = renderer.getMapTileFromCenter();

			final int additionalTilesNeededToLeftOfCenter;
			final int additionalTilesNeededToRightOfCenter;
			final int additionalTilesNeededToTopOfCenter;
			final int additionalTilesNeededToBottomOfCenter;

			Pixel upperLeftCornerOfCenterMapTile = renderer
					.getUpperLeftCornerOfCenterMapTileInScreen(null);

			final int centerMapTileScreenLeft = upperLeftCornerOfCenterMapTile
					.getX();
			final int centerMapTileScreenTop = upperLeftCornerOfCenterMapTile
					.getY();

			final int centerMapTileScreenRight = centerMapTileScreenLeft
					+ tileSizePx;
			final int centerMapTileScreenBottom = centerMapTileScreenTop
					+ tileSizePx;

			int[] additionalTiles = bufferStrategy
					.calculateBufferAroundCenterTile((int) mapWidth,
							(int) mapHeight, tileSizePx,
							centerMapTileScreenLeft, centerMapTileScreenRight,
							centerMapTileScreenTop, centerMapTileScreenBottom);

			additionalTilesNeededToLeftOfCenter = additionalTiles[0];
			additionalTilesNeededToRightOfCenter = additionalTiles[1];
			additionalTilesNeededToTopOfCenter = additionalTiles[2];
			additionalTilesNeededToBottomOfCenter = additionalTiles[3];

			final int mapTileUpperBound = bufferStrategy
					.calculateTileUpperBound(zoomLevel);
			final int[] mapTileCoords = new int[] {
					centerMapTileCoords[MAPTILE_LATITUDE_INDEX],
					centerMapTileCoords[MAPTILE_LONGITUDE_INDEX] };

			final int size = (additionalTilesNeededToBottomOfCenter
					+ additionalTilesNeededToTopOfCenter + 1)
					* (additionalTilesNeededToRightOfCenter
							+ additionalTilesNeededToLeftOfCenter + 1);
			Tile[] tiles = new Tile[size];
			int cont = 0;

			Extent maxExtent = renderer.getExtent();

			if (mapWidth < mapRenderer.getMAPTILE_SIZEPX()
					|| mapHeight == mapRenderer.getMAPTILE_SIZEPX()) {
				maxExtent = this.mapRenderer.getExtentIncludingCenter(extent
						.getCenter(), zoomLevel, new Point(mapRenderer
						.getOriginX(), mapRenderer.getOriginY()));
				double eWidth = maxExtent.getWidth();
				double eHeight = maxExtent.getHeight();
				mapWidth = Math.max(
						(int) (eWidth / mapRenderer.resolutions[zoomLevel]), 1);
				mapHeight = Math
						.max(
								(int) (eHeight / mapRenderer.resolutions[zoomLevel]),
								1);
			}

			final Extent viewExtent = ViewPort.calculateExtent(renderer
					.getCenter(), renderer.resolutions[zoomLevel], mapWidth,
					mapHeight);

			double[] coords = new double[] {
					maxExtent.getLefBottomCoordinate().getX(),
					maxExtent.getLefBottomCoordinate().getY() };
			final int[] leftBottom = renderer.toPixels(coords);

			coords = new double[] { maxExtent.getRightTopCoordinate().getX(),
					maxExtent.getRightTopCoordinate().getY() };
			final int[] rightTop = renderer.toPixels(coords);

			IRect r = this.getContext().getRectangle();
			r.set(leftBottom[0], rightTop[1], rightTop[0], leftBottom[1]);

			boolean process = true;

			for (int y = -additionalTilesNeededToTopOfCenter; y <= additionalTilesNeededToBottomOfCenter; y++) {
				for (int x = -additionalTilesNeededToLeftOfCenter; x <= additionalTilesNeededToRightOfCenter; x++) {
					process = true;
					if (cancellable.getCanceled()) {
						this.mTileProvider.clearPendingQueue();
						WorkQueue.getInstance().clearPendingTasks();
						context.sendMessage(callBackHandler, new BaseEvent(
								"Tile provider canceled"),
								GeoUtils.TILEPROVIDER_CANCELED);
						return;
					}

					if (viewExtent.intersect(maxExtent)) {
						final int tileLeft = ViewPort.mTouchMapOffsetX
								+ centerMapTileScreenLeft + (x * tileSizePx);
						final int tileTop = ViewPort.mTouchMapOffsetY
								+ centerMapTileScreenTop + (y * tileSizePx);

						process &= r.intersects(tileLeft, tileTop, tileLeft
								+ tileSizePx / 2, tileTop + tileSizePx / 2);
					}

					if (process) {
						mapTileCoords[MAPTILE_LATITUDE_INDEX] = GeoMath.mod(
								centerMapTileCoords[MAPTILE_LATITUDE_INDEX]
										+ renderer.isTMS() * y,
								mapTileUpperBound);

						mapTileCoords[MAPTILE_LONGITUDE_INDEX] = GeoMath.mod(
								centerMapTileCoords[MAPTILE_LONGITUDE_INDEX]
										+ x, mapTileUpperBound);

						final String tileURLString = renderer.getTileURLString(
								mapTileCoords, zoomLevel);

						final int[] tile = new int[] { mapTileCoords[0],
								mapTileCoords[1] };

						final int tileLeft = ViewPort.mTouchMapOffsetX
								+ centerMapTileScreenLeft + (x * tileSizePx);
						final int tileTop = ViewPort.mTouchMapOffsetY
								+ centerMapTileScreenTop + (y * tileSizePx);

						if (this.checkIntersects(tile, renderer, zoomLevel)) {
							final Tile t = new Tile(tileURLString, tile,
									new Pixel(tileLeft, tileTop), zoomLevel,
									renderer.getNAME(), cancellable, null);
							if (tileSorter == null) {
								this.mTileProvider.getMapTile(t.clone(),
										cancellable);
							} else {
								if (cont < tiles.length)
									tiles[cont] = t;
							}
						} else {
							context
									.sendMessage(
											callBackHandler,
											new BaseEvent(
													"Tile skipped because tile does not intersect with Geometry: " + tileURLString),
											GeoUtils.TILEPROVIDER_SKIPPED);
						}
					} else {
						context.sendMessage(callBackHandler, new BaseEvent(
								"Tile skipped"), GeoUtils.TILEPROVIDER_SKIPPED);
					}
					cont++;
				}
			}

			if (tileSorter != null) {
				tileSorter.sortTiles(tiles, this.centerPixelX,
						this.centerPixelY, tileSizePx);
				final int length = tiles.length;
				Tile temp;
				for (int j = 0; j < length; j++) {
					temp = tiles[j];
					if (temp != null) {
						// This should notify the callback Handlers when the
						// tile is
						// cached
						this.mTileProvider
								.getMapTile(temp.clone(), cancellable);
					}
				}
			}

			tiles = null;

		} catch (Exception e) {
			log.log(Level.FINE, "onDraw", e);
		}
	}

	protected boolean checkIntersects(final int[] tile,
			final MapRenderer renderer, final int zoomLevel) {
		boolean result = true;

		if (this.tileBufferIntersector != null) {
			double originX = renderer.getOriginX();
			double originY = renderer.getOriginY();	
			
			final Extent tileExtent = renderer.getTileExtent(tile[1],
					tile[0], renderer.resolutions[zoomLevel], originX, originY);
			result = tileBufferIntersector.intersects(tileExtent);
		}

		return result;
	}

	/**
	 * 
	 * @return The total number of tiles to download according to the current
	 *         TileGenerator attributes and IBufferStrategy
	 */
	public long calculateTotalNumTilesToDownload() {
		final MapRenderer renderer = this.mapRenderer;
		final int tileSizePx = renderer.getMAPTILE_SIZEPX();

		long totalTiles = 0;

		for (int zoomLevel = this.fromZoomLevel; zoomLevel <= this.toZoomLevel; zoomLevel++) {
			renderer.setZoomLevel(zoomLevel);

			if (this.downloadAllExtent) {
				this.recalcWidthHeightOfAreaToDownload(zoomLevel);
			}

			final int additionalTilesNeededToLeftOfCenter;
			final int additionalTilesNeededToRightOfCenter;
			final int additionalTilesNeededToTopOfCenter;
			final int additionalTilesNeededToBottomOfCenter;

			Pixel upperLeftCornerOfCenterMapTile = renderer
					.getUpperLeftCornerOfCenterMapTileInScreen(null);

			final int centerMapTileScreenLeft = upperLeftCornerOfCenterMapTile
					.getX();
			final int centerMapTileScreenTop = upperLeftCornerOfCenterMapTile
					.getY();

			final int centerMapTileScreenRight = centerMapTileScreenLeft
					+ tileSizePx;
			final int centerMapTileScreenBottom = centerMapTileScreenTop
					+ tileSizePx;

			int[] additionalTiles = bufferStrategy
					.calculateBufferAroundCenterTile((int) mapWidth,
							(int) mapHeight, tileSizePx,
							centerMapTileScreenLeft, centerMapTileScreenRight,
							centerMapTileScreenTop, centerMapTileScreenBottom);

			additionalTilesNeededToLeftOfCenter = additionalTiles[0];
			additionalTilesNeededToRightOfCenter = additionalTiles[1];
			additionalTilesNeededToTopOfCenter = additionalTiles[2];
			additionalTilesNeededToBottomOfCenter = additionalTiles[3];

			totalTiles += (additionalTilesNeededToBottomOfCenter
					+ additionalTilesNeededToTopOfCenter + 1)
					* (additionalTilesNeededToRightOfCenter
							+ additionalTilesNeededToLeftOfCenter + 1);

		}

		return totalTiles;
	}

	/**
	 * The TileProvider
	 * 
	 * @return
	 */
	public TileProvider getTileProvider() {
		return mTileProvider;
	}

	/**
	 * Sets the tile provider
	 * 
	 * @param tileProvider
	 */
	public void setTileProvider(TileProvider tileProvider) {
		mTileProvider = tileProvider;
	}

	private IContext getContext() throws BaseException {
		if (this.context == null)
			context = CompatManager.getInstance().getRegisteredContext();
		return context;
	}

	/**
	 * Sets the bufferstrategy
	 * 
	 * @param bufferStrategy
	 */
	public void setBufferStrategy(IBufferStrategy bufferStrategy) {
		this.bufferStrategy = bufferStrategy;
	}

	/**
	 * The buffer strategy
	 * 
	 * @return
	 */
	public IBufferStrategy getBufferStrategy() {
		return bufferStrategy;
	}

	/**
	 * Sets the TileSorter
	 * 
	 * @param tileSorter
	 */
	public void setTileSorter(ITileSorter tileSorter) {
		this.tileSorter = tileSorter;
	}

	/**
	 * The TileSorter
	 * 
	 * @return
	 */
	public ITileSorter getTileSorter() {
		return tileSorter;
	}

	public void destroy() {
		this.mTileProvider.destroy();
	}

	public void clearQueue(boolean cancel) {
		if (cancel)
			this.cancellable.setCanceled(true);
		this.mTileProvider.clearPendingQueue();
	}

	public void onLowMemory() {
		clearQueue(true);
	}
}
