/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2009 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2009.
 *   author Alberto Romeu aromeu@prodevelop.es 
 *   author Ruben Blanco rblanco@prodevelop.es 
 *   
 */

package es.prodevelop.tilecache.renderer;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.http.util.ByteArrayBuffer;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

import es.prodevelop.geodetic.utils.conversion.ConversionCoords;
import es.prodevelop.gvsig.mini.geom.Extent;
import es.prodevelop.gvsig.mini.geom.Pixel;
import es.prodevelop.gvsig.mini.geom.Point;
import es.prodevelop.gvsig.mini.map.CapabilitiesTags;
import es.prodevelop.gvsig.mini.map.ViewPort;
import es.prodevelop.gvsig.mini.projection.TileConversor;
import es.prodevelop.gvsig.mobile.fmap.proj.CRSFactory;
import es.prodevelop.gvsig.mobile.fmap.proj.Projection;

/**
 * MapRednerer for TileMapServices
 * 
 * @author aromeu
 * @author rblanco
 * 
 */
public class TMSRenderer extends MapRenderer {

	private final static Logger logger = Logger
			.getLogger("TMSRenderer");
	Vector tempRes = new Vector();
	static String encoding = "UTF-8";
	private Extent currentExtent;
	private double currentRes;

	protected TMSRenderer(final String[] aBaseUrl, final String aName,
			final String aImageFilenameEnding, final int aZoomMax,
			final int aZoomMin, final int aTileSizePX, final int type,
			final Point origin, final Extent maxExtent, final String SRS,
			final double[] resolutions, Extent offExtent) {
		super(aBaseUrl, aName, aImageFilenameEnding, aZoomMax, aZoomMin,
				aTileSizePX, type, origin, maxExtent, SRS, resolutions, offExtent);
	}

	/**
	 * Connects to the URL to retrieve the info of the layer in the
	 * TileMapService and returns an instance of a TMSRenderer ready to consume
	 * tiles
	 * 
	 * @param url
	 *            The URL of a layer of a Tile Map Service
	 * @param layerName
	 *            The name of the layer (visible for users)
	 * @return A TMSRenderer
	 */
	public static TMSRenderer getTMSRenderer(final String url, String layerName) {
		TMSRenderer t = TMSRenderer.getTMSRenderer(url, layerName, null, 1, 0,
				256, MapRenderer.TMS_RENDERER, null, null, null, null, null);

		t.tryParseTileSet(url);
		return t;
	}

	/**
	 * Loads the properties of a TMSRenderer formatted into a String[] @see
	 * Layers and layers.txt
	 * 
	 * @param props
	 *            The properties of the TMSRenderer
	 * @param layerName
	 *            The name of the layer
	 * @return A TMSRenderer
	 */
	public static TMSRenderer loadProperties(final String[] props,
			String layerName) {
		try {
			int index = props[0].indexOf("[");
			Extent offExtent = null;
			String offlineLayerName = "";
			if (index != -1) {
				try {
					String ex = props[0].substring(index + 1, props[0].length() - 1);
					if (ex != null && ex.compareTo("") != 0) {
						String[] parts = ex.split(MapRenderer.NAME_SEPARATOR);						
						offExtent = Extent.parseString(parts[0], ":");
						offlineLayerName = parts[1];
					}
				} catch (Exception e) {
					
				}
				
			}
				
			int type = Integer.valueOf(props[0].substring(0, 1)).intValue();
			String[] url = props[1].split(MapRenderer.URL_SEPARATOR);
			String format = props[2];
			int minZoomLevel = Integer.valueOf(props[4]).intValue();
			int maxZoomLevel = Integer.valueOf(props[3]).intValue();
			int tileSize = Integer.valueOf(props[5]).intValue();
			double minX = Double.valueOf(props[8]).doubleValue();
			double minY = Double.valueOf(props[9]).doubleValue();
			double maxX = Double.valueOf(props[10]).doubleValue();
			double maxY = Double.valueOf(props[11]).doubleValue();
			double originX = Double.valueOf(props[6]).doubleValue();
			double originY = Double.valueOf(props[7]).doubleValue();
			String SRS = props[12];

			String[] resol = props[13].split(":");
			final int length = resol.length;
			double[] resolutions = new double[length];
			for (int i = 0; i < length; i++) {
				try {
					resolutions[i] = Double.valueOf(resol[i]).doubleValue();
				} catch (NumberFormatException e) {
					
				}				
			}

			TMSRenderer renderer = new TMSRenderer(url,
					layerName, format, maxZoomLevel, minZoomLevel, tileSize,
					type, new Point(originX, originY), new Extent(minX, minY,
							maxX, maxY), SRS, resolutions, offExtent);

			return renderer;
		} catch (Exception e) {
			logger.log(Level.SEVERE, "getTMSRenderer: ", e);
			return null;
		}
	}

	/**
	 * Instantiates a TMSRenderer
	 * 
	 * @param props
	 *            if props contains the name of the layer and the URL, connects
	 *            to the Tile Map Service to retrieve the TMSRenderer
	 *            properties. If props contains all the properties simply
	 *            instantiates the TMSRenderer
	 * @param layerName
	 *            The name of the layer
	 * @return A TMSRenderer
	 */
	public static TMSRenderer getTMSRenderer(final String[] props,
			String layerName) {

		try {
			if (props.length == 2) {
				return TMSRenderer.getTMSRenderer(props[1], layerName);
			}
			return TMSRenderer.loadProperties(props, layerName);
		} catch (Exception e) {
			logger.log(Level.SEVERE, "getTMSRenderer: ", e);
			return null;
		}
	}

	public static TMSRenderer getTMSRenderer(final String aBaseUrl,
			final String aName, final String aImageFilenameEnding,
			final int aZoomMax, final int aZoomMin, final int aTileSizePX,
			final int type, final Point origin, final Extent maxExtent,
			final String SRS, final double[] resolutions, Extent offExtent) {

		TMSRenderer r = new TMSRenderer(new String[] { aBaseUrl }, aName,
				aImageFilenameEnding, aZoomMax, aZoomMin, aTileSizePX, type,
				origin, maxExtent, SRS, resolutions, offExtent);
		return r;
	}

	@Override
	public String getTileURLString(final int[] tileID, final int zoomLevel) {
		String s = new StringBuilder().append(this.getBASEURL()).append(
				zoomLevel).append("/").append(tileID[1]).append("/").append(
				tileID[0]).append(".").append(this.getIMAGE_FILENAMEENDING())
				.toString();
		// logger.debug(s);
		return s;
	}

	public double[] getResolutions() {
		return this.resolutions;
	}

	/**
	 * parses TileSet tag
	 */
	public final void tryParseTileSet(String url) {
		URL parseURL = null;
		URLConnection urlconnec = null;
		InputStream is = null;
		BufferedInputStream bis = null;
		ByteArrayBuffer baf = null;

		try {
			/* Define the URL we want to load data from. */
			parseURL = new URL(url);
			/* Open a connection to that URL. */
			urlconnec = parseURL.openConnection();
			/* Define InputStreams to read from the URLConnection. */
			is = urlconnec.getInputStream();
			bis = new BufferedInputStream(is);
			/* Read bytes to the Buffer until there is nothing more to read(-1). */
			baf = new ByteArrayBuffer(50);
			int current = 0;
			while ((current = bis.read()) != -1) {
				baf.append((byte) current);
			}

			int tag;
			final KXmlParser kxmlParser = createParser(baf.toByteArray());
			kxmlParser.nextTag();

			if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
				kxmlParser.require(KXmlParser.START_TAG, null,
						CapabilitiesTags.TILE_MAP);
				tag = kxmlParser.nextTag();
				while (tag != KXmlParser.END_DOCUMENT) {
					switch (tag) {

					case KXmlParser.START_TAG:
						if (kxmlParser.getName()
								.compareTo(CapabilitiesTags.SRS) == 0) {
							kxmlParser.next();
							SRS = kxmlParser.getText();
						} else if (kxmlParser.getName().compareTo(
								CapabilitiesTags.BOUNDINGBOX) == 0) {
							double minx = 0;
							double maxx = 0;
							double miny = 0;
							double maxy = 0;
							String value = "";

							try {

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.MINX);
								if (value != null) {
									minx = Double.parseDouble(value);
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.MAXX);
								if (value != null) {
									maxx = Double.parseDouble(value);
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.MINY);
								if (value != null) {
									miny = Double.parseDouble(value);
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.MAXY);
								if (value != null) {
									maxy = Double.parseDouble(value);
								}

								this.setExtent(new Extent(minx, miny, maxx,
										maxy));
								// if (ext.area() > 0)
								// this.BBOX = ext;
								// else
								// SRS = "";

							} catch (Exception ex) {
								logger.log(Level.SEVERE, "", ex);
							}

						} else if (kxmlParser.getName().compareTo(
								CapabilitiesTags.ORIGIN) == 0) {
							double x = 0;
							double y = 0;
							String value = "";

							try {

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.X);
								if (value != null) {
									this.setOriginX(Double.parseDouble(value));
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.Y);
								if (value != null) {
									this.setOriginY(Double.parseDouble(value));
								}
							} catch (Exception ex) {
								logger.log(Level.SEVERE, "", ex);
							}

						} else if (kxmlParser.getName().compareTo(
								CapabilitiesTags.TILE_FORMAT) == 0) {
							String value = "";

							try {
								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.WIDTH);
								if (value != null) {
									this.setMAPTILE_SIZEPX(Integer
											.parseInt(value));
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.HEIGHT);
								if (value != null) {
									this.setMAPTILE_SIZEPX(Integer
											.parseInt(value));
								}

								value = kxmlParser.getAttributeValue("",
										CapabilitiesTags.EXTENSION);
								if (value != null) {
									this.setIMAGE_FILENAMEENDING(value);
								}

							} catch (Exception ex) {
								logger.log(Level.SEVERE, "", ex);
							}

						} else if (kxmlParser.getName().compareTo(
								CapabilitiesTags.TILE_SETS) == 0) {
							this.parseTileSets(kxmlParser);
						}

						break;
					case KXmlParser.END_TAG:
						break;
					case KXmlParser.TEXT:
						break;
					}

					tag = kxmlParser.next();
				}

				kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
			}

		} catch (XmlPullParserException parser_ex) {
			logger.log(Level.SEVERE, "", parser_ex);
		} catch (IOException ioe) {
			logger.log(Level.SEVERE, "", ioe);
		} finally {
			if (tempRes != null) {
				final int length = tempRes.size();
				resolutions = new double[length];
				for (int i = 0; i < length; i++) {
					resolutions[i] = Double.valueOf(
							tempRes.elementAt(i).toString()).doubleValue();
				}
				this.setZOOM_MAXLEVEL(length - 1);
			}
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
					logger.log(Level.SEVERE, "", e);
				}
			}

			if (bis != null) {
				try {
					bis.close();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

			if (baf != null) {
				baf.clear();
			}
		}
	}

	/**
	 * parse TileSet tag
	 * 
	 * @param parser
	 * @throws java.io.IOException
	 * @throws org.xmlpull.v1.XmlPullParserException
	 */
	public final void parseTileSets(final KXmlParser parser)
			throws IOException, XmlPullParserException {
		int currentTag;
		boolean end = false;

		parser.require(KXmlParser.START_TAG, null, CapabilitiesTags.TILE_SETS);
		currentTag = parser.next();

		// tileSets = new Hashtable();
		while (!end) {
			switch (currentTag) {
			case KXmlParser.START_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.TILE_SET) == 0) {
					String url = "";
					String res = "";
					try {
						url = parser.getAttributeValue("",
								CapabilitiesTags.HREF);
						res = parser.getAttributeValue("",
								CapabilitiesTags.UNITS_PER_PIXEL);
						tempRes.add(res);
					} catch (final Exception e) {
						logger.log(Level.SEVERE, "", e);
					}
				}
			case KXmlParser.END_TAG:
				if (parser.getName().compareTo(CapabilitiesTags.TILE_SETS) == 0) {
					end = true;
				}

				break;
			case KXmlParser.TEXT:
				break;
			}

			if (!end) {
				currentTag = parser.next();
			}
		}
		// parser.require(KXmlParser.END_TAG, null,
		// CapabilitiesTags.CAPABILITY);
	}

	public static final KXmlParser createParser(final byte[] f) {
		KXmlParser parser = null;
		ByteArrayInputStream br;
		byte[] buffer;
		try {
			br = new ByteArrayInputStream(f);
			buffer = new byte[100];
			br.read(buffer);
			String string = new String(buffer);

			final int a = string.toLowerCase().indexOf("<?xml");
			if (a != -1) {
				string = string.substring(a, string.length());
			}
			final StringBuffer st = new StringBuffer(string);
			final String searchText = "encoding=\"";
			final int index = st.toString().indexOf(searchText);
			if (index > -1) {
				st.delete(0, index + searchText.length());
				encoding = st.toString().substring(0,
						st.toString().indexOf("\""));
			}
		} catch (final Exception e) {
			logger.log(Level.SEVERE, "", e);
			return null;
		}
		parser = new KXmlParser();
		try {
			br = new ByteArrayInputStream(f);
			buffer = new byte[100];
			br.read(buffer);
			String string = new String(buffer);
			int a = string.toLowerCase().indexOf("<?xml");
			if (a != -1) {
				string = string.substring(a, string.length());
				parser.setInput(br, encoding);
			} else // end patch
			{
				parser.setInput(br, encoding);
			}
			br.reset();
		} catch (Exception ex) {
			logger.log(Level.SEVERE, "", ex);
		} finally {
			buffer = null;
			try {
				if (br != null) {
					br.close();
				}
			} catch (final Exception e) {
			}
		}
		return parser;
	}

	@Override
	public int isTMS() {
		return -1;
	}

	@Override
	public String toString() {
		StringBuffer layer = new StringBuffer();
		try {
			final int length = resolutions.length;
			StringBuffer res = new StringBuffer();
			for (int i = 0; i < length; i++) {
				res.append(resolutions[i]);
				if (i != length - 1) {
					res.append(":");
				}
			}
			
			String offExtent = "";
			if (this.isOffline()) 
				offExtent = this.getOfflineExtent().toString(":");

			layer.append(this.getFullNAME()).append(";")
			.append(this.getType()).append("[").append(offExtent).append(MapRenderer.NAME_SEPARATOR).append(this.getOfflineLayerName()).append("]").append(",")
			.append(this.getURLtoString()).append(",")
			.append(this.getIMAGE_FILENAMEENDING()).append(",")
			.append(this.getZOOM_MAXLEVEL()).append(",")
			.append(this.getZoomMinLevel()).append(",")
			.append(this.getMAPTILE_SIZEPX()).append(",")
			.append(getOriginX()).append(",")
			.append(getOriginY()).append(",")
			.append(getExtent().toString()).append(",")
			.append(SRS).append(",").append(res.toString());
			return layer.toString();

		} catch (Exception e) {
			logger.log(Level.SEVERE, "toString: ", e);
			return null;
		}
	}

	@Override
	public double[] transformCenter(String toCRS) {
		try {
			return ConversionCoords.reproject(center.getX(), center.getY(),
					CRSFactory.getCRS(this.getSRS()), CRSFactory.getCRS(toCRS));
		} catch (Exception e) {
			logger.log(Level.SEVERE, "transformCoords: ", e);
			return null;
		}
	}

	@Override
	public void setCenter(double x, double y) {
		this.center = new Point(x, y);		
		currentExtent = ViewPort.calculateExtent(this.getCenter(), currentRes,
				ViewPort.mapWidth, ViewPort.mapHeight);
		double[] latLonCenter = transformCenter("EPSG:4326");
		this.getLatLonCenter().setCoordinates(latLonCenter);
	}

	@Override
	public Point getCenter() {
		return new Point(this.center.getX(), this.center.getY());
	}

	public Pixel getUpperLeftCornerOfCenterMapTileInScreen(final Pixel reuse) {
		final Pixel out = (reuse != null) ? reuse : new Pixel(0, 0);
		// Pixel tile = TileConversor.latLonToTileG(this.getCenter().getX(),
		// this
		// .getCenter().getY(), Map.vp.getDist1Pixel());
		Pixel tile = TileConversor.metersToTile(this.getCenter().getX(), this
				.getCenter().getY(),
				this.getResolutions()[this.getZoomLevel()], -this.getOriginX(),
				-this.getOriginY());
		// Extent e1 = TileConversor.tileBoundsG(tile.getX(), tile.getY(),
		// Map.vp.getDist1Pixel());
		Extent e1 = TileConversor.tileMeterBounds(tile.getX(), tile.getY(),
				this.getResolutions()[this.getZoomLevel()]);
		Extent e2 = ViewPort.calculateExtent(
				new es.prodevelop.gvsig.mini.geom.Point(
						this.getCenter().getX(), this.getCenter().getY()), this
						.getResolutions()[this.getZoomLevel()],
				ViewPort.mapWidth, ViewPort.mapHeight);
		int[] centerTopLeft = ViewPort.fromMapPoint(new double[] {
				e1.getMinX(), e1.getMaxY() }, e2.getMinX(), e2.getMaxY(), this
				.getResolutions()[this.getZoomLevel()]);
		out.setX(centerTopLeft[0]);
		out.setY(centerTopLeft[1]);
		return out;
	}

	public int[] getMapTileFromCenter() {
		// Pixel tile = TileConversor.latLonToTileG(this.getCenter().getX(),
		// this
		// .getCenter().getY(), Map.vp.getDist1Pixel());
		Pixel tile = TileConversor.metersToTile(this.getCenter().getX(), this
				.getCenter().getY(),
				this.getResolutions()[this.getZoomLevel()], -this.getOriginX(),
				-this.getOriginY());
		return new int[] { tile.getY(), tile.getX() };
	}

	@Override
	public double[] fromPixels(int[] pxy) {
		pxy[0] -= ViewPort.mTouchMapOffsetX;
		pxy[1] -= ViewPort.mTouchMapOffsetY;
		final double res = this.resolutions[this.getZoomLevel()];
		return ViewPort.toMapPoint(pxy, (int) ViewPort.mapWidth,
				(int) ViewPort.mapHeight, this.getCenter(), res);
	}

	@Override
	public Pixel getCenterE6() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public int[] toPixels(double[] pxy) {
		final double res = currentRes;
		final Extent e = currentExtent;
		int[] result = ViewPort
				.fromMapPoint(pxy, e.getMinX(), e.getMaxY(), res);
		result[0] += ViewPort.mTouchMapOffsetX;
		result[1] += ViewPort.mTouchMapOffsetY;
		return result;
	}
	
	@Override
	public void setZoomLevel(int z) {
		super.setZoomLevel(z);
		currentRes = this.resolutions[this.getZoomLevel()];
		currentExtent = ViewPort.calculateExtent(this.getCenter(), currentRes,
				ViewPort.mapWidth, ViewPort.mapHeight);
	}

	@Override
	public void centerOnBBox() {
		if (this.isOffline()) {
			this.center = this.getOfflineExtent().getCenter();
			this.setZoomLevel((this.getZOOM_MAXLEVEL() - this.getZoomMinLevel()) / 2);
		} else {
			this.center = this.getExtent().getCenter();
			this.setZoomLevel(this.getZoomMinLevel());
		}
	}

	@Override
	public String buildQuery(Extent e, int zoomLevel, double resolution) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Extent getExtentIncludingCenter(Point center, int zoomLevel,
			Point origin) {
		Pixel tile = TileConversor.metersToTile(center.getX(), center.getY(),
				this.getResolutions()[zoomLevel], -this.getOriginX(), -this
						.getOriginY());
		Extent ex = TileConversor.tileMeterBounds((int) tile.getX(), (int) tile
				.getY(), this.getResolutions()[this.getZoomLevel()]);

		// TODO Auto-generated method stub
		return ex;
	}

	@Override
	public StringBuffer getQuadkey(final int tileX, final int tileY,
			final int zoomLevel) {
		try {
			if (tileX == -1) {
				return null;
			} else {
				return TileConversor.tileXYToQuadKey(tileX, tileY, zoomLevel);
			}
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public Pixel getTileNumber(Extent e, int zoomLevel, double resolution,
			Projection proj) {
		Pixel tile = TileConversor.metersToTile(e.getCenter().getX(), e
				.getCenter().getY(), resolution, -this.getOriginX(), -this
				.getOriginY());
		return tile;
	}

	@Override
	public Extent getTileExtent(int tileX, int tileY, double resolution, double originX,
			double originY) {
		Extent e1 = TileConversor.tileMeterBounds(tileX, tileY,
				resolution, originX, originY);
		return e1;
	}

	@Override
	public int[] toPixelsFromLonLat(double[] pxy) {
		final double res = ViewPort.LAT_LON_RES[this.getZoomLevel()];
		Extent e = ViewPort.calculateExtent(this.getLatLonCenter(), res,
				ViewPort.mapWidth, ViewPort.mapHeight);
		int[] result = ViewPort
				.fromMapPoint(pxy, e.getMinX(), e.getMaxY(), res);
		result[0] += ViewPort.mTouchMapOffsetX;
		result[1] += ViewPort.mTouchMapOffsetY;
		return result;
	}

	/**
	 * @return the currentExtent
	 */
	@Override
	public Extent getCurrentExtent() {
		return currentExtent;
	}

	/**
	 * @return the currentRes
	 */
	@Override
	public double getCurrentRes() {
		return currentRes;
	}
}
